/*----------------------------------------------------------------------- 

File Name: ArrayFactories.c

<lalVerbatim file="ArrayFactoriesCV">
Revision: $Id: ArrayFactories.m4,v 1.3 2000/11/11 22:43:34 jolien Exp $
</lalVerbatim>

-------------------------------------------------------------------------*/

/* <lalLaTeX>

\subsection{Module \texttt{ArrayFactories.c}}
\label{ss:ArrayFactories.c}

Create/destroy $\langle\mbox{datatype}\rangle$Array objects. 

\subsubsection*{Prototypes}
\vspace{0.1in}
\input{ArrayFactoriesD}
\index{\texttt{LALZCreateArray()}}
\index{\texttt{LALCCreateArray()}}
\index{\texttt{LALDCreateArray()}}
\index{\texttt{LALSCreateArray()}}
\index{\texttt{LALI2CreateArray()}}
\index{\texttt{LALI4CreateArray()}}
\index{\texttt{LALI8CreateArray()}}
\index{\texttt{LALU2CreateArray()}}
\index{\texttt{LALU4CreateArray()}}
\index{\texttt{LALU8CreateArray()}}
\index{\texttt{LALCreateArray()}}
\index{\texttt{LALZDestroyArray()}}
\index{\texttt{LALCDestroyArray()}}
\index{\texttt{LALDDestroyArray()}}
\index{\texttt{LALSDestroyArray()}}
\index{\texttt{LALI2DestroyArray()}}
\index{\texttt{LALI4DestroyArray()}}
\index{\texttt{LALI8DestroyArray()}}
\index{\texttt{LALU2DestroyArray()}}
\index{\texttt{LALU4DestroyArray()}}
\index{\texttt{LALU8DestroyArray()}}
\index{\texttt{LALDestroyArray()}}

\subsubsection*{Description}

The \texttt{CreateArray} family of functions create a
$\langle\mbox{datatype}\rangle$\texttt{Array} of the appropriate dimensions.

The \texttt{DestroyArray} family of functions return the storage allocated by
the \texttt{CreateArray} functions to the system.

\subsubsection*{Algorithm}

\subsubsection*{Uses}
\begin{verbatim}
LALMalloc()
LALFree()
\end{verbatim}

\subsubsection*{Notes}

\vfill{\footnotesize\input{ArrayFactoriesCV}}

</lalLaTeX> */

#include <string.h>
#include "LALStdlib.h"
#include "AVFactories.h"

/* <lalVerbatim file="ArrayFactoriesNRCSID"> */
NRCSID( ARRAYFACTORIESC, "$Id: ArrayFactories.m4,v 1.3 2000/11/11 22:43:34 jolien Exp $" );
/* </lalVerbatim> */
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALZCreateArray ( LALStatus *status, COMPLEX16Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALZCreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( COMPLEX16Array * ) LALMalloc ( sizeof( COMPLEX16Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( COMPLEX16 * ) LALMalloc( arrayDataSize*sizeof( COMPLEX16 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALZDestroyArray ( LALStatus *status, COMPLEX16Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALZDestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALCCreateArray ( LALStatus *status, COMPLEX8Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALCCreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( COMPLEX8Array * ) LALMalloc ( sizeof( COMPLEX8Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( COMPLEX8 * ) LALMalloc( arrayDataSize*sizeof( COMPLEX8 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALCDestroyArray ( LALStatus *status, COMPLEX8Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALCDestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALDCreateArray ( LALStatus *status, REAL8Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALDCreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( REAL8Array * ) LALMalloc ( sizeof( REAL8Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( REAL8 * ) LALMalloc( arrayDataSize*sizeof( REAL8 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALDDestroyArray ( LALStatus *status, REAL8Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALDDestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALSCreateArray ( LALStatus *status, REAL4Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALSCreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( REAL4Array * ) LALMalloc ( sizeof( REAL4Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( REAL4 * ) LALMalloc( arrayDataSize*sizeof( REAL4 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALSDestroyArray ( LALStatus *status, REAL4Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALSDestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALI2CreateArray ( LALStatus *status, INT2Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALI2CreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( INT2Array * ) LALMalloc ( sizeof( INT2Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( INT2 * ) LALMalloc( arrayDataSize*sizeof( INT2 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALI2DestroyArray ( LALStatus *status, INT2Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALI2DestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALI4CreateArray ( LALStatus *status, INT4Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALI4CreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( INT4Array * ) LALMalloc ( sizeof( INT4Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( INT4 * ) LALMalloc( arrayDataSize*sizeof( INT4 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALI4DestroyArray ( LALStatus *status, INT4Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALI4DestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALI8CreateArray ( LALStatus *status, INT8Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALI8CreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( INT8Array * ) LALMalloc ( sizeof( INT8Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( INT8 * ) LALMalloc( arrayDataSize*sizeof( INT8 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALI8DestroyArray ( LALStatus *status, INT8Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALI8DestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALU2CreateArray ( LALStatus *status, UINT2Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALU2CreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( UINT2Array * ) LALMalloc ( sizeof( UINT2Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( UINT2 * ) LALMalloc( arrayDataSize*sizeof( UINT2 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALU2DestroyArray ( LALStatus *status, UINT2Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALU2DestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALU4CreateArray ( LALStatus *status, UINT4Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALU4CreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( UINT4Array * ) LALMalloc ( sizeof( UINT4Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( UINT4 * ) LALMalloc( arrayDataSize*sizeof( UINT4 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALU4DestroyArray ( LALStatus *status, UINT4Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALU4DestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALU8CreateArray ( LALStatus *status, UINT8Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALU8CreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( UINT8Array * ) LALMalloc ( sizeof( UINT8Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( UINT8 * ) LALMalloc( arrayDataSize*sizeof( UINT8 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALU8DestroyArray ( LALStatus *status, UINT8Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALU8DestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALCreateArray ( LALStatus *status, REAL4Array **array, UINT4Vector *dimLength ) 
{ /* </lalVerbatim> */
  UINT4 arrayDataSize = 1;
  UINT4 numDims;
  UINT4 dim;

  INITSTATUS (status, "LALCreateArray", ARRAYFACTORIESC);   
  ATTATCHSTATUSPTR (status);

  /* make sure arguments are sane */

  ASSERT (array,             status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (!*array,           status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT (dimLength,         status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->data,   status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (dimLength->length, status,
          AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  numDims = dimLength->length;

  /* loop over dimensions to compute total size of array data */

  for (dim = 0; dim < numDims; ++dim)
  {
    arrayDataSize *= dimLength->data[dim];
  }

  ASSERT (arrayDataSize, status, AVFACTORIESH_ELENGTH, AVFACTORIESH_MSGELENGTH);

  /* allocate memory for structure */

  *array = ( REAL4Array * ) LALMalloc ( sizeof( REAL4Array ) );
  ASSERT (*array, status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);

  (*array)->dimLength = NULL;
  (*array)->data      = NULL;

  /* allocate dimLength field and copy information there */

  LALU4CreateVector (status->statusPtr, &(*array)->dimLength, numDims);
  if (status->statusPtr->statusCode)
  {
    LALFree (*array);
    *array = NULL;
    RETURN (status);   /* this returns a recursive error status code (-1) */
  }
  memcpy ((*array)->dimLength->data, dimLength->data, numDims*sizeof(UINT4));
  
  /* allocate storage */

  (*array)->data = ( REAL4 * ) LALMalloc( arrayDataSize*sizeof( REAL4 ) );

  if (!(*array)->data)
  {
    /* try to free memory                                               */
    /* this should ALWAYS work so we can use the CHECKSTATUSPTR() macro */
    LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
    CHECKSTATUSPTR (status);  /* if this fails, there is a memory leak  */

    LALFree (*array);
    *array = NULL;
    ABORT (status, AVFACTORIESH_EMALLOC, AVFACTORIESH_MSGEMALLOC);
  }

  DETATCHSTATUSPTR (status);
  RETURN (status);
}
















/* <lalVerbatim file="ArrayFactoriesD"> */
void LALDestroyArray ( LALStatus *status, REAL4Array **array )
{ /* </lalVerbatim> */
  INITSTATUS (status, "LALDestroyArray", ARRAYFACTORIESC);	
  ATTATCHSTATUSPTR (status);
      
  ASSERT (array,          status, AVFACTORIESH_EVPTR, AVFACTORIESH_MSGEVPTR);
  ASSERT (*array,         status, AVFACTORIESH_EUPTR, AVFACTORIESH_MSGEUPTR);
  ASSERT ((*array)->data, status, AVFACTORIESH_EDPTR, AVFACTORIESH_MSGEDPTR);

  /* Free allocated storage */

  LALU4DestroyVector (status->statusPtr, &(*array)->dimLength);
  CHECKSTATUSPTR (status);

  LALFree ((*array)->data); /* free allocated data */
  LALFree (*array);	    /* free array struct itself */
  *array = NULL;	    /* make sure we don't point to freed struct */

  DETATCHSTATUSPTR (status);
  RETURN (status);
}

